"""\
flake8-string package.

Checks for literal string usage: quotes, format, f-strings
"""

plugin_prefix = 'LIT'

quote_checker_prefix = f'{plugin_prefix}0'
raw_checker_prefix = f'{plugin_prefix}1'
